
// matches.h
// Copyright 2015 Matthew Rickard
// This file is part of dep

// dep is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#ifndef MATCHES_H
#define MATCHES_H

#include <string>
#include <vector>

#include <regex>

// Note: <regex> is unimplemented/experimental as of g++ 4.8.5 (Amazon Linux AMI 1)
// It was implemented and released in g++ 4.9.0
//
// https://stackoverflow.com/questions/12530406/is-gcc-4-8-or-earlier-buggy-about-regular-expressions

struct Pattern {
  Pattern(const char *pattern, std::regex_constants::syntax_option_type patternType);
  std::string s;
  std::regex re;
};

struct StartPattern {
  StartPattern(const char *pattern, std::regex_constants::syntax_option_type patternType, int flags);
  std::string s; // The full string
  std::regex re; // Just the part with the wildcards etc
  int useRegex; // Boolean: There are wildcards etc
  //const char *p;
  int howFar;
  int flags;
};

bool matches(const std::string &s, const Pattern &pattern);
bool contains(const std::string &s, const Pattern &pattern, int *pos = 0, int *len = 0);
bool contains(const std::string &s, const StartPattern &pattern, int *pos = 0, int *len = 0);

// Other bits+pieces

#define STARTSWITH_NONE  0
#define STARTSWITH_FIXED 1

int startsWith(int flags, const char *s, const char *start);
int startsWith(int flags, const char *s, const char *start, int startlen, int *len/*, int *fixed*/);

int patternStartsWith(const char *s, const char *start, const StartPattern &pattern,
  int startlen, int *len);

int endsWith(int flags, const char *s, const char *end);
int endsWith(int flags, const char *s, const char *end, int endlen);

// Lexer (Lexical Analyzer/Lexical Scanner)

struct ScanResult {
  int token;
  const char *str;
  int len;
};

int operator == (const ScanResult &a, const ScanResult &b);

#define LEXER_DISCARD 1

struct Token {
  Token(const char *regexp, const char *code, int flags);
  Pattern pattern;
  const char *code;
  int flags;
};

struct Lexer {
  int addToken(const char *regexp, const char *code, int flags);
  const char *scan(const char *s, ScanResult *result);
  std::vector<Token> exprVector;
  void scanAll(const char *input, std::vector<ScanResult> &scanResult, std::string *keyString);
};

#endif

